//     Author: Maurizio Loreti, aka MLO or (HAM) I3NOO
//     Work:   University of Padova - Department of Physics
//             Via F. Marzolo, 8 - 35131 PADOVA - Italy
//     Phone:  +39 (049) 827-7216   FAX: +39 (049) 827-7102
//     EMail:  loreti@pd.infn.it
//     WWW:    http://www.pd.infn.it/~loreti/mlo.html
//
// -------------------------------------------------------------------
// Definition of the static data and the methods declared in
// synctime.hh; see the comments there for more.
// -------------------------------------------------------------------
// $Id: synctime.cxx,v 1.6 2004/02/15 10:07:17 loreti Exp $
// -------------------------------------------------------------------

#include <locale>
#include <stdexcept>
#include <sstream>

extern "C" {
  #include <errno.h>
  #include <stdio.h>
  #include <stdlib.h>
#ifdef NEED_TZ
  #include <langinfo.h>
#endif // NEED_TZ
  #include <sys/types.h>
  #include <netdb.h>
  #include <signal.h>
}

#include "synctime.hh"

using std::cerr;
using std::ostringstream;
using std::runtime_error;
using std::string;

// Static data for the timeSocket and unixTime classes

const int          timeSocket::_timePort = 37;
const unsigned int timeSocket::_timeOut  = 10;

const time_t unixTime::unixEpoch = 2208988800UL;

#ifdef NEED_TZ
  string unixTime::printFormat;
#endif // NEED_TZ

namespace {
  void alarmHandler(
    int sigNum
  ) {
    // Called if, in an UDP connection, too much time has elapsed

    if (sigNum == SIGALRM) {
      cerr << "recvfrom: connection timed out\n";
      exit(EXIT_FAILURE);
    }
  }
}

bool aToAddr(
  const char *         address,
  struct sockaddr_in * pSaddr
) {
  // Converts the constant char array "address" (the trusted host
  // address) to the data internally stored in the "in_addr"
  // structure.  The procedure tries first to decode "address" a
  // numbers-and-dots  identifier (calling inet_aton), then as an
  // alphanumeric identifier (calling gethostbyname); then returns
  // 'true' if a conversion was successful, 'false' if not.

  if (inet_aton(address, &(pSaddr->sin_addr)) != 0) return true;

  struct hostent * host = gethostbyname(address);
  if (host) {
    if (host->h_addrtype != AF_INET) {
      cerr << address << ": not an AF_INET address\n";
      return false;
    }
    memcpy(&(pSaddr->sin_addr), host->h_addr_list[0], host->h_length);
    return true;
  }

  return false;
}

timeSocket::timeSocket(
  const string & host,
  socketType     type,
  bool           quiet
) : _name(host), _type(type), _quiet(quiet), _socket(-1) {
  // Stores into the private in_addr structure the network data of the
  // remote host, and of its well-known time port

  if (!aToAddr(host.c_str(), &_host)) {
    string s(host);
    s += ": invalid host name";
    throw runtime_error(s);
  }

  _host.sin_family = AF_INET;
  _host.sin_port   = htons(_timePort);
}

void timeSocket::getConnection() {
  // Gets a socket file descriptor from the operating system

  switch (_type) {
    case UDP:
      _socket = socket(AF_INET, SOCK_DGRAM, 0);
      break;

    case TCP:
      _socket = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
      break;

    default:
      ostringstream s;
      s << "Wrong socket type = " << _type;
      throw runtime_error(s.str());
  }

  if (_socket < 0) {
    string s("socket: ");
    s += strerror(errno);
    throw runtime_error(s);
  }

  if (!_quiet) std::cout << "Connecting via " << _type << " to "
                         << inet_ntoa(_host.sin_addr) << " on port "
                         << ntohs(_host.sin_port) << " ...\n";
}

int32_t timeSocket::getTime() {
  // Queries the remote server to obtain the value of the POSIX time

  if (_socket == -1) getConnection();

  int32_t           t32;
  socklen_t         bytes_read;
  int               this_read;
  char *            buf = reinterpret_cast<char *>(&t32);
  struct sockaddr * pSA = reinterpret_cast<struct sockaddr *>(&_host);

  switch (_type) {
    case UDP:
      // UDP protocol.  Establishes a signal handler for SIGALRM;
      // sends an empty message to the time port of the remote host;
      // and reads back the 32 bits of data containing the POSIX
      // time.  The SIGALRM is user to set a timeout.

      bytes_read = sizeof(_host);

      if (signal(SIGALRM, alarmHandler) == SIG_ERR) {
        string s("signal: ");
        s += strerror(errno);
        throw runtime_error(s);
      }

      if (sendto(_socket, 0, 0, 0, pSA, bytes_read)) {
        string s("sendto: ");
        s += strerror(errno);
        throw runtime_error(s);
      }

      alarm(_timeOut);
      if (recvfrom(_socket, buf, sizeof(t32), 0, pSA,
                   &bytes_read) != sizeof(t32)) {
        string s;

        if (bytes_read > 0) {
          s  = "recvfrom: ";
          s += strerror(errno);
        } else {
          s  = "EOF received from ";
          s += _name;
        }
        throw runtime_error(s);
      }
      alarm(0);
      break;

    case TCP:
      // TCP protocol; opens a connection with the time port of the
      // remote host, and reads 32 bits of data (with the same
      // meaning)

      if (connect(_socket, pSA, sizeof(_host))) {
        string s("connect: ");
        s += strerror(errno);
        throw runtime_error(s);
      }

      for (bytes_read = 0;  bytes_read < sizeof(t32);  ) {
        do {
          this_read = read(_socket, buf, sizeof(t32) - bytes_read);
        } while (this_read < 0  &&  errno == EINTR);

        if (this_read < 0) {
          string s("read: ");
          s += strerror(errno);
          throw runtime_error(s);
        } else if (this_read == 0) {
          break;
        } else {
          bytes_read += this_read;
          buf        += this_read;
        }
      }

      if (bytes_read != sizeof(t32)) {
        throw runtime_error("read: input error");
      }
      break;
  }

  return ntohl(t32);
}

string unixTime::toString() const {
  // Converts in a printable string (according to the current locale)
  // the date/time value stored in _time

  std::locale l;
  static bool message = false;

  try {
    l = std::locale("");
  } catch (std::exception & e) {
    if (!message) {
      message = true;
      cerr << "*** EXCEPTION - " << e.what() << " - C locale assumed\n";
    }
    l = std::locale::classic();
  }

  ostringstream s;
  s.imbue(l);

  struct tm *        pTm = localtime(&_time);

#ifndef NEED_TZ
  const char         fmt[]  = "%c";
  const char * const fmtEnd = fmt + sizeof(fmt);
  std::use_facet< std::time_put<char> >(l).put(
    std::ostreambuf_iterator<char>(s), s, s.fill(), pTm, fmt, fmtEnd);
#else
// The above lines do not work as expected; the "%c" specification in
// fmt[], when used by strftime(3) (with g++ 3.3.2), does not obtain
// for some locales (notably for the "C" and the "POSIX" locale) the
// information about the time zone.

  static const char zone[] = " %Z";
  if (printFormat.empty()) {
    printFormat = nl_langinfo(D_T_FMT);
    if (printFormat.find(zone+1) == string::npos) printFormat += zone;
  }
  std::use_facet< std::time_put<char> >(l).put(
    std::ostreambuf_iterator<char>(s), s, s.fill(), pTm,
    printFormat.c_str(), printFormat.c_str()+printFormat.size());
#endif // NEED_TZ

  return s.str();
}
